/*
 * SSL Smurf
 *
 * The SSL Smurf acts as a local HTTP proxy that dumps all packets
 * it passes to and fro to standard output. The "smurf" part 
 * consists of a capability to convert HTTP requests for certain
 * hosts to HTTPS (SSL). This feature is used to dump the clear
 * text of HTTPS web applications.
 *
 * Example:
 *
 * ./sslsmurf -P proxy:8080 -h www.secureserver.org
 *
 * The sslsmurf opens a local port and acts as an HTTP proxy, so point
 * your web browser to use the sslsmurf as its proxy. If you surf to
 * http://www.secureserver.org/bla then the SSL smurf converts the
 * request to an SSL connection. With this command the SSL smurf uses
 * the downstream proxy proxy:8080 for its connections to the Internet.
 *
 * (c) Copyright 2004 Jos Visser (Muppet) <josv@osp.nl>
 *
 * This code is licensed under the GNU General Public License (GPL).
 * See the file LICENSE for more information...
 *
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <stdarg.h>
#include <pthread.h>
#include <fnmatch.h>

#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/x509.h>
#include <openssl/pem.h>
#include <openssl/ssl.h>
#include <openssl/err.h>

#define LINELEN		2048	/* Maximum length of a single line (for my_printf) */
#define BUFSIZE		65536	/* Size of the internal buffer used for data transfer */
#define MAX_HDRS	128	/* Maximum number of headers in a HTTP request */
#define VERSION		"0.9.3"	/* Guess what */

#define WHITESPACE	" \r\n\t"

#define MARKER		"========================================================================"
#define REPLY200	"HTTP/1.0 200 Connection established\r\n\r\n"
#define REPLY500	"HTTP/1.0 500 There is an error in smurf country\r\n\r\n"

/*
 * Three macro's used in checking the return codes of SSL connections
 */
#define CHK_NULL(x,s) if ((x)==NULL) { my_perror(s); flush_output(); exit (1); }
#define CHK_SSL(err) if ((err)==-1) { my_perror("SSL error"); flush_output(); ERR_print_errors_fp(stderr); exit(2); }

/*
 * This structure describes an incoming HTTP request broken down to its
 * component parts
 */
struct request {
	char *buf;
	int  smurf_this;
	char *cmd;
	char *url;
	char *version;
	char *protocol;
	char *host;
	char *port;
	char *file;
	char *connreq;
};

/*
 * This structure contains output lines for the threadsafe buffered output feature
 */
struct output_chunk {
	char *data;
	struct output_chunk *next;
};

/*
 * This structure contains a hostname for which HTTP requests must
 * be smurfed to HTTPS requests
 */
struct host_to_smurf {
	char *host;
	struct host_to_smurf *next;
};

int 			no_threading;	/* Don't use multi threading */
struct host_to_smurf 	*root;		/* The root of the list of hosts to smurf */
char			*proxy_host;	/* The hostname of the downstream proxy to connect to */
int			proxy_port;	/* Guess what? */
int			port=6066;	/* The local port to accept requests on */
SSL_CTX			*ctx;		/* The SSL context we use to create new SSL connections */
int			verbose_flag;	/* Guess what? */
int			stop;		/* Indicator to stop the sslsmurf */
char			*client_cert;	/* Client certificate file to be used (PEM) */
char			*client_privkey;/* Client private key file to be used (PEM) */
pthread_key_t		buffer_key;	/* The pthreads key for the thread global buffer */
pthread_key_t		output_key;	/* The pthreads key for the list of output chunks */
pthread_mutex_t		output_mutex	/* Mutex to guard output to standard out */ 
				=PTHREAD_MUTEX_INITIALIZER;

/*
 * Two simple macros to determine the largest/smallest item 
 */
#ifndef MAX
#define MAX( x, y )     ( ( (x)>(y) ) ? (x) : (y) )
#endif

#ifndef MIN
#define MIN( x, y )     ( ( (x)>(y) ) ? (y) : (x) )
#endif

/*
 * My own printf...
 */
void my_printf(char *s, ...) {
	char *buf=(char *)malloc(BUFSIZE);
	struct output_chunk *work=(struct output_chunk *)malloc(sizeof(struct output_chunk));
	struct output_chunk **root=(struct output_chunk **)pthread_getspecific(output_key);

	vsprintf(buf,s, (char *) &s + sizeof(s));
	work->data=strdup(buf);
	work->next=*root;
	*root=work;
	free(buf);
}

/*
 * My own puts
 */
void my_puts(char *s) {
	my_printf("%s\n",s);
}

/*
 * Write a line without terminating newline
 */
void my_putline(char *s) {
	my_printf("%s",s);
}

/*
 * My own perror
 */
void my_perror(char *s) {
	my_printf("Error; %s: %s\n",s,strerror(errno));
}

/*
 * Flush the buffered output chunks to stdout (recursive routine)
 */
void flush_output2(struct output_chunk *work) {
	char *s;
	int newline=1;
	pthread_t me=pthread_self();

	if (!work)
		return;

	flush_output2(work->next);

	s=work->data;

	/*
	while (*s) {
		if (newline) {
			printf("[%08x] ",(unsigned int)me);
			newline=0;
		}

		putchar(*s);

		if (*s=='\n')
			newline=1;

		++s;
	}
	*/
	printf("%s",s);

	free(work->data);
	free(work);
}

/*
 * Flush the buffered output chunks to stdout
 */
void flush_output() {
	struct output_chunk **root=(struct output_chunk **)pthread_getspecific(output_key);

	if (!*root)
		return;

	pthread_mutex_lock(&output_mutex);
	flush_output2(*root);
	fflush(stdout);
	pthread_mutex_unlock(&output_mutex);
	*root=0;
}

/*
 * Hexdump a buffer of data to the screen. The output is a mixed
 * hex/ascii display in the best tradition of hexdumps all throughout
 * history.
 */
void hexdump(char *buf, int n, char *direction, int from, int to) {
	unsigned char *p=(unsigned char *)buf;
	int count;
	int i;
	int tot=0;
	char line[LINELEN];
	char s[LINELEN];
	time_t	now=time(0);

	if (direction) {
		my_printf(ctime(&now));
		my_printf("Block of %d (0x%x) bytes going %s (%0x -> %0x):\n",n,n,direction,from,to);
	}

	while (n>0) {
		/*
		 * Each line contains 16 bytes of output
		 */
		count=MIN(n,16);

		/*
		 * Print the offset of the current line
		 */
		sprintf(line, "%08X: ",tot);

		/*
		 * Print the hexadecimal values
		 */
		for (i=0; i<count; i++) {
			sprintf(s,"%02X ",p[i]);
			strcat(line,s);
		}

		/*
		 * If the current line contains less than 16
		 * bytes, fill out with spaces
		 */
		for (i=0; i<(16-count); i++)
			strcat(line,"   ");

		strcat(line,"   ");

		/*
		 * Now print the ASCII characters for all
		 * bytes that are in the displayable
		 * character range
		 */
		for (i=0; i<count; i++)
			if (p[i]>31 && p[i]<128) {
				sprintf(s,"%c",p[i]);
				strcat(line,s);
			} else
				strcat(line,".");

		n-=count;
		tot+=count;
		p+=count;

		/*
		 * And on to the next line of output
		 */
		my_puts(line);
	}

	my_printf("\n");
	flush_output();
}


/*
 * Read one line of data from a socket. A line is terminated by a
 * newline character. The result is stored in the global buf.
 * This routine is wildly inefficient, squandering resources as
 * if there is no tomorrow...
 */
char *readline(int sd)
{
        char   	c = 0;
        int    	i = 0;
	char	*buf=(char *)pthread_getspecific(buffer_key);
        char   	*p = buf;

        /*
         * Read one character at a time into buf, until a newline is
         * encountered.
         */
        while ( c != 10 && i < BUFSIZE - 1 ) {
                if( recv( sd, &c ,1 ,0 ) <= 0) {
                        my_perror( "Socket read error (readline)" );
			*p=0;
			return buf;
                }

                *p = c;
                p++;
                i++;
        }

	/*
	 * If the last character read is not a newline, there
	 * was a buffer overflow. Dump the buffer for good
	 * measure and continue
	 */
	if (c!=10) {
		my_puts("*** Buffer overflow!!!");
		hexdump(buf,BUFSIZE,0,0,0);
	}

	/*
	 * Ensure the buffer is terminated with a null
	 * character. This can be helpful later on...
	 */
        *p = 0;

	return buf;
}

/*
 * Open a TCP connection to the given host/port. Returns
 * the socket. If the connection can not be made the
 * sslsmurf terminates...
 */
int open_connection(char *host, int port) {
	int 			sd;
	struct sockaddr_in 	sa;
	struct hostent 		*he;
	
	/*
	 * Locate the IP address of the host
	 */
	he=gethostbyname(host);

	/*
	 * If the IP address can not be found, terminate
	 * the program with an error
	 */
	if (!he) {
		my_printf("Error resolving hostname: %s\n",host);

		return -1;
	}

	/*
	 * Create a new socket
	 */
	sd=socket (AF_INET,SOCK_STREAM,0);
 
 	/*
	 * Set up the address (IP/port) to connect to
	 */
	memset (&sa, '\0', sizeof(sa));
  	sa.sin_family=AF_INET;
	memcpy( &sa.sin_addr.s_addr, he->h_addr, he->h_length);
  	sa.sin_port= htons(port);

	/*
	 * Give a nice output message
	 */
	my_printf( "Connecting to %s(%d.%d.%d.%d):%d\n", host, he->h_addr[0] & 255,
                                he->h_addr[1] & 255,
                                he->h_addr[2] & 255,
                                he->h_addr[3] & 255,port );
  
  	/*
	 * And connect the socket. If the connection fails stop
	 * with an error
	 */
  	if (connect(sd,(struct sockaddr*) &sa,sizeof(sa))<0) {
		my_perror("connect() failed");

		return -1;
	}

	return sd;
}

/*
 * Copy a block of data from one socket descriptor to another. A true
 * return code signifies EOF on the from socket descriptor.
 */
int copy(int from, int to, char *direction) {
	int 		n;
	char		*buf=(char *)pthread_getspecific(buffer_key);

	/* 
	 * Read a buffer from the source socket
	 */
	if ( ( n = read( from, buf, BUFSIZE ) ) < 0 ) {
		my_perror( "Socket read error (copy)" );
		flush_output();
		exit( 1 );
	}

	/*
	 * If we have read 0 bytes, there is an EOF on src
	 */
	if( n==0 )
		return 1;

	/* 
	 * Read a buffer from the source socket
	 */
	hexdump(buf,n,direction,from,to);

	/*
	 * Write the buffer to the destination socket
	 */
	if ( write( to, buf, n ) != n ) {
		my_perror( "Socket write error" );
		flush_output();
		exit( 1 );
	}

	/*
	 * We're not yet at EOF
	 */
	return 0;
}

/*
 * Execute a loop of copying data between the browser and the
 * web server or downstream proxy.
 */
void cpio(int sock1, int sock2) {
	fd_set	readfds;
	fd_set	writefds;
	fd_set	exceptfds;
	int	max_fd;

	/*
	 * Find the biggest file descriptor for select()
	 */
	max_fd = MAX( sock1, sock2 );

	/*
	 * We're never interested in sockets being available for write.
	 */
	FD_ZERO( &writefds );

	/*
	 * Only diamonds are forever :-)
	 */
	while( 1==1 )
	{
		/*
		 * Clear the interesting socket sets
		 */
		FD_ZERO( &readfds );
		FD_ZERO( &exceptfds );

		/*
		 * We want to know whether both sockets are ready for read
		 */
		FD_SET( sock1, &readfds );
		FD_SET( sock2, &readfds );

		/*
		 * And we want to know about exceptional conditions on either
		 * socket
		 */
		FD_SET( sock1, &exceptfds );
		FD_SET( sock2, &exceptfds );

		/*
		 * Wait until something happens on one of the registered
		 * sockets/files
		 */
		if ( select( max_fd + 1, &readfds, &writefds, &exceptfds, 0 ) < 0 ) {
			my_perror("select error");
			flush_output();
			exit(1);
		}

		/*
		 * Is sock1 ready for read? If so, copy a block of data
		 * from sock1 to sock2. Or else if sock2 
		 * is ready for read, copy a block of data from
		 * sock2 to sock1. Otherwise an exceptional condition
		 * is flagged and the copying is terminated
		 */
		if ( FD_ISSET( sock1, &readfds ) ) {
			if ( copy( sock1, sock2, "in" ) )
				break;
		}
		else if( FD_ISSET( sock2, &readfds ) ) {
			if( copy(sock2,sock1, "out" ) )
				break;
		}
		else {
			my_perror( "Exceptional condition" );
			break;
		}
	}
}

/*
 * Should HTTP connections for this host be smurfed? Runs
 * through the list of hosts to be smurfed and returns true
 * if this hose is in the list.
 */
int smurf_it(char *host) {
	struct host_to_smurf *work=root;

	while (work && fnmatch(work->host,host,0)!=0)
		work=work->next;

	return work!=0;
}

/*
 * Break an HTTP request into its atomic parts...
 */
struct request *breakdown_request(char *request) {
	struct request *req=(struct request *)malloc(sizeof(struct request));

	/*
	 * Make a private copy of the original request string
	 */
	req->buf=strdup(request);

	/*
	 * Remember that a request looks like:
	 *
	 * GET http://host:port/file HTTP/1.0
	 *             or
	 * CONNECT host:port HTTP/1.0
	 */

	/*
	 * First take the command
	 */
	req->cmd=strtok(req->buf,WHITESPACE);

	/* 
	 * Then the URL
	 */
	req->url=strtok(0,WHITESPACE);

	/*
	 * And the protocol version
	 */
	req->version=strtok(0,WHITESPACE);

	/*
	 * If the command is a CONNECT, there is no protocol or file
	 * Also find the host:port part.
	 */
	if (strcasecmp(req->cmd,"CONNECT")==0) {
		req->protocol=0;
		req->host=req->url;
		req->file=0;
	} else {
		req->protocol=strtok(req->url,":");
		req->host=strtok(0,"/");
		req->file=strtok(0,"");
	}

	/*
	 * Split the host:port in a separate host and port field
	 */
	req->host=strtok(req->host,":");
	req->port=strtok(0,"");

	/*
	 * If the port is missing the port is 443
	 * (this is to get the smurfing right
	 */
	if (!req->port)
		req->port="443";

	/*
	 * Should this request be smurfed?
	 * This is true if the host must be smurfed and it is a normal GET
	 * request. If the host is in the to-be-smurfed list but the command
	 * is already a CONNECT request then no smurfing should take place
	 */
	req->smurf_this=smurf_it(req->host) && (strcasecmp(req->cmd,"CONNECT")!=0);

	/*
	 * Give some extra info if requested
	 */

	if (verbose_flag) {
		my_printf("Request is: %s",request);
		my_printf("cmd=%s; protocol=%s; host=%s; port=%s; file=%s; version=%s\n",
			req->cmd, req->protocol, req->host, req->port, req->file,
			req->version);
		my_printf("This request will%s be smurfed.\n\n",req->smurf_this?"":" not");
	}

	/*
	 * If this request must be smurfed prepare the CONNECT request we should
	 * present to the downstream proxy (if there is one)
	 */
	if (req->smurf_this) {
		// CONNECT host:port version  
		// 12345678    9     A      BCD
		req->connreq=(char *)malloc(13+strlen(req->host)+strlen(req->port)+strlen(req->version));
		sprintf(req->connreq,"CONNECT %s:%s %s\r\n",req->host,req->port,req->version);
	} else
		req->connreq=0;

	return req;
}

/*
 * Free the request structure
 */
void free_request(struct request *req) {
	free(req->buf);
	
	if (req->connreq)
		free(req->connreq);

	free(req);
}

/*
 * Process an unsmurfed request. This means connecting to the web server or the
 * downstream proxy and passing the request.
 */
void process_unsmurfed(struct request *req, int sock, char *request, char *headers[]) {
	int sock_out;
	int i;

	/*
	 * Connect to the web server or the downstream proxy
	 */
	if (!proxy_host)
		sock_out=open_connection(req->host,atoi(req->port));
	else
		sock_out=open_connection(proxy_host,proxy_port);

	/*
	 * Did the open of the output connection succeed?
	 */
	if (sock_out>=0) {
		my_printf("Connection from: %x -> %x\n",sock,sock_out);

		/*
	 	 * If the command is a CONNECT, we have just fulfilled our
	 	 * duty. Just connect the in/out sockets in a tunnel fashion
	 	 * and we're done
	 	 */
		if (strcasecmp(req->cmd,"CONNECT")==0 && !proxy_host) {
			my_puts("Not re-issuing request\n");
			write(sock,REPLY200,strlen(REPLY200));
		} else {
			/*
		 	 * Otherwise we must re-issue the request to the
		 	 * web server or downstream proxy
		 	 */

			if (verbose_flag)
				my_puts("Re-issuing request:\n");

			write(sock_out,request,strlen(request));

			if (verbose_flag)
				my_putline(request);

			for (i=0; headers[i]; i++) {
				write(sock_out,headers[i],strlen(headers[i]));

				if (verbose_flag)
					my_putline(headers[i]);
			}
	
			/*
		 	 * After the request we need to send a blank line
		 	 */
			write(sock_out,"\r\n",2);

			if (verbose_flag)
				my_puts("");
		}

		/*
	 	 * Now copy blocks of data to and fro
	 	 */
		cpio(sock,sock_out);

		/*
	 	 * And close the output socket
	 	 */
		close(sock_out);
	} else {
		/*
		 * The open failed, let's return an error message
		 */
		write(sock,REPLY500,strlen(REPLY500));
	}
}

/*
 * Report an SSL error
 */
void ssl_error(SSL *ssl, int n, char *operation) {
	int	ssl_err;

	ssl_err=SSL_get_error(ssl,n);
	my_printf("SSL error %d on %s of SSL socket %x\n",ssl_err,operation,ssl);
	ERR_print_errors_fp(stderr);
	flush_output();
}


/*
 * Copy a block of data from a normal socket to an SSL socket. A true
 * return code signifies EOF on the from socket descriptor.
 */
int copy_n2ssl(int from, SSL *to, char *direction) {
	int 	n;
	int	ssl_err;
	char	*buf=(char *)pthread_getspecific(buffer_key);

	/* 
	 * Read a buffer from the source socket
	 */
	if ( ( n = read( from, buf, BUFSIZE ) ) < 0 ) {
		my_perror( "Socket read error (copy_n2ssl)" );
		flush_output();
		return 1;
	}

	/*
	 * If we have read 0 bytes, there is an EOF on src
	 */
	if( n==0 )
		return 1;

	/*
	 * Dump the block we have read to the screen
	 */
	hexdump(buf,n,direction,from,(int)to);

	/*
	 * Write the buffer to the destination SSL socket
	 */
	if ( (ssl_err=SSL_write( to, buf, n )) != n ) {
		ssl_error(to,ssl_err,"write");
		return 1;
	}

	/*
	 * We're not yet at EOF
	 */
	return 0;
}

/*
 * Copy a block of data from an SSL socket to a normal socket. A true
 * return code signifies EOF on the from socket descriptor.
 */
int copy_ssl2n(SSL *from, int to, char *direction) {
	int 	n;		
	int	ssl_err;
	char	*buf=(char *)pthread_getspecific(buffer_key);

	/* 
	 * Read a buffer from the source socket
	 */
	if ( ( n = SSL_read( from, buf, BUFSIZE ) ) < 0 ) {
		ssl_err=SSL_get_error(from,n);

		if (ssl_err==SSL_ERROR_WANT_READ)
			return 0;

		ssl_error(from,n,"read");
		return 1;
	}

	/*
	 * If we have read 0 bytes, there is an EOF on src
	 */
	if( n==0 )
		return 1;

	/* 
	 * Read a buffer from the source socket
	 */
	hexdump(buf,n,direction,(int)from,to);

	/*
	 * Write the buffer to the destination socket
	 */
	if ( write( to, buf, n ) != n ) {
		my_perror( "Socket write error" );
		flush_output();
		return 1;
	}

	/*
	 * We're not yet at EOF
	 */
	return 0;
}

/*
 * Move into a loop of copying data between the browser and the
 * remote HTTPS server (or through a downstream proxy).
 */
void ssl_cpio(int sock1, int sock2, SSL *sock2_ssl)
{
	fd_set	readfds;
	fd_set	writefds;
	fd_set	exceptfds;
	int	max_fd;

	/*
	 * Find the biggest file descriptor for select()
	 */
	max_fd = MAX( sock1, sock2 );

	/*
	 * We're never interested in sockets being available for write.
	 */
	FD_ZERO( &writefds );

	/*
	 * Only diamonds are forever :-)
	 */
	while( 1==1 )
	{
		/*
		 * Clear the interesting socket sets
		 */
		FD_ZERO( &readfds );
		FD_ZERO( &exceptfds );

		/*
		 * We want to know whether both sockets are ready for read
		 */
		FD_SET( sock1, &readfds );
		FD_SET( sock2, &readfds );

		/*
		 * And we want to know about exceptional conditions on either
		 * stdin, stdout or the tunnel
		 */
		FD_SET( sock1, &exceptfds );
		FD_SET( sock2, &exceptfds );

		/*
		 * Wait until something happens on one of the registered
		 * sockets/files
		 */
		if ( select( max_fd + 1, &readfds, &writefds, &exceptfds, 0 ) < 0 ) {
			my_perror("select error");
			flush_output();
			exit(1);
		}

		/*
		 * Is sock1 ready for read? If so, copy a block of data
		 * from sock1 to sock2. Or else if sock2 
		 * is ready for read, copy a block of data from
		 * sock2 to sock1. Otherwise an exceptional condition
		 * is flagged and the copying is terminated.
		 */
		if ( FD_ISSET( sock1, &readfds ) ) {
			if ( copy_n2ssl( sock1, sock2_ssl, "out" ) )
				break;
		}
		else if( FD_ISSET( sock2, &readfds ) ) {
			if( copy_ssl2n(sock2_ssl, sock1, "in" ) )
				break;
		}
		else {
			my_perror( "Exceptional condition" );
			break;
		}
	}
}

/*
 * Process a request that must be smurfed. This means that the downstream connection
 * should be an SSL connection.
 */
void process_smurfed(struct request *req, int sock, char *request, char *headers[]) {
	int	err;
	int	sock_out;
	SSL	*ssl;
	X509	*server_cert;
	int	i;
	char	*inheaders[MAX_HDRS];
	char	*rc;
	char	*buf=(char *)pthread_getspecific(buffer_key);

	/*
	 * If we do not have a downstream proxy we create a socket connection
	 * to the web server
	 */
	if (!proxy_host)
		sock_out=open_connection(req->host,atoi(req->port));
	else {
		/*
		 * If we do have a downstream proxy we must open a
		 * connection to the proxy and execute the CONNECT protocol
		 */
		sock_out=open_connection(proxy_host,proxy_port);

		if (verbose_flag) {
			my_puts("Opened connection to downstream proxy");
			my_puts("CONNECT request going out:");
			my_puts("-------------------------");
		}

		write(sock_out,req->connreq,strlen(req->connreq));

		if (verbose_flag)
			my_putline(req->connreq);

		/*
		 * Send out the proxy-authorization headers from the
		 * original request
		 */
		for (i=0; headers[i]; i++)
			if (strncasecmp(headers[i],"proxy-a",7)==0) {
				write(sock_out,headers[i],strlen(headers[i]));

				if (verbose_flag)
					my_putline(headers[i]);
			}

		/*
		 * And send the end-of-request marker
		 */
		write(sock_out,"\r\n",2);

		if (verbose_flag)
			my_puts("");

		/*
	 	* Now read the response to the CONNECT request
	 	*/
		if (verbose_flag) {
			my_puts("Reply to the CONNECT (coming in):");
			my_puts("--------------------------------");
		}

		*buf=0;
		i=0;

		while (!(*buf=='\r' && *(buf+1)=='\n')) {
			inheaders[i]=strdup(readline(sock_out));

			if (verbose_flag)
				my_putline(inheaders[i]);
			i++;
		}

		i--;
		free(inheaders[i]);
		inheaders[i]=0;

		/*
		 * Check whether the response to our CONNECT was a 
		 * return code 200 (OK). If it is not we forward the
		 * response to the browser and we're ready.
		 */
		strcpy(buf,inheaders[0]);
		strtok(buf,WHITESPACE);
		rc=strtok(0,WHITESPACE);

		if (strcmp(rc,"200")!=0) {
			for (i=0; inheaders[i]; i++) {
				write(sock,inheaders[i],strlen(inheaders[i]));
				free(inheaders[i]);
			}

			close(sock_out);

			return;
		}

		/*
		 * Free the headers from the CONNECT response
		 */
		for (i=0; inheaders[i]; i++)
			free(inheaders[i]);
	}
		
	/*
	 * Did the connection to the downstream node fail?
	 */
	if (sock_out<0) 
		return;

	/* 
	 * At this point we have a TCP socket to the downstream web server.
	 * Create a new SSL socket and start negotiation.
	 */
	ssl = SSL_new (ctx);                         
	CHK_NULL(ssl,"SSL_new(ctx)");    
	SSL_set_fd (ssl, sock_out);
	err = SSL_connect (ssl);
	CHK_SSL(err);

	/*
	 * Give some helpful data from the SSL socket
	 */
	my_puts("SSL socket info:");
	my_puts("---------------");
	my_printf ("Cipher used: %s\n", SSL_get_cipher (ssl));
	server_cert = SSL_get_peer_certificate (ssl); 
	CHK_NULL(server_cert,"SSL_get_peer_certificate(ssl)");
	my_puts ("Server certificate:");
	X509_NAME_oneline (X509_get_subject_name (server_cert),buf,BUFSIZE);
	my_printf ("  Subject: %s\n", buf);
	X509_NAME_oneline (X509_get_issuer_name  (server_cert),buf,BUFSIZE);
	my_printf ("  Issuer: %s\n\n", buf);
	X509_free(server_cert);

	printf("Connection: from %x -> %x (SSL; orig is %x)\n",sock,(int)ssl,sock_out);

	/*
	 * Now we have the SSL socket set up. Forward the request to the target
	 * web server (headers and all, but not any proxy- related headers.
	 */
	SSL_write(ssl,request,strlen(request));
	
	for (i=0; headers[i]; i++)
		if (strncasecmp(headers[i],"proxy-",6)!=0)
			SSL_write(ssl,headers[i],strlen(headers[i]));

	/*
	 * And of course do not forget the end-of-request marker
	 */
	SSL_write(ssl,"\r\n",2);

	/*
	 * Copy blocks of data to/from the SSL socket
	 */
	ssl_cpio(sock,sock_out,ssl);

	/* 
	 * Clean up, close, forget, release
	 */
	SSL_shutdown(ssl);
	close(sock_out);
	SSL_free (ssl);
}

/*
 * Process incoming connections. If this function returns
 * "true", the program must be terminated
 */
void *process(void *arg) {
	int			sock=(int)arg;
	char 			*request;
	char			*newreq;
	char 			*headers[MAX_HDRS];
	int 			i=0;
	int 			j;
	struct request		*req;
	time_t			now=time(0);
	char			*buf=(char *)alloca(BUFSIZE);
	struct output_chunk 	*root=0;

	/*
	 * Check whether the buffer allocation went allright
	 */
	if (!buf) {
		my_perror("Could not create buffer!");
		close(sock);
		return 0;
	}

	/*
	 * Store the buffer in the thread local storage keyspace
	 */
	if (pthread_setspecific(buffer_key,buf)!=0) {
		my_perror("Could not store buffer in TLS");
		close(sock);
		return 0;
	}
	
	/*
	 * Store the output chunk list root in the thread local storage keyspace
	 */
	if (pthread_setspecific(output_key,&root)!=0) {
		my_perror("Could not store output chunk list root in TLS");
		close(sock);
		return 0;
	}

	/*
	 * Read the request line from the incoming socket
	 */
	request=strdup(readline(sock));

	/*
	 * If the request is a "quit", terminate the program (nicely)
	 */
	if (strncasecmp(request,"QUIT",4)==0) {
		my_puts("QUIT instruction accepted");
		stop=1;
		close(sock);
		return 0;
	}

	/*
	 * Create the header on the screen
	 */
	my_puts(MARKER);
	my_printf(ctime(&now));
	my_puts(MARKER);
	my_putline(request);
	
	/*
	 * Read the request headers from the stream. If there are
	 * more than MAX_HDRS headers an error is given and the
	 * processing of this request is terminated.
	 */
	while (1==1) {
		if (i==MAX_HDRS) {
			my_puts("Too many header lines in request!");
			my_printf("This is the request: %s\n",request);
			my_puts("These are the headers so far:");

			for (j=0; j<i; j++) {
				my_putline(headers[i]);
				free(headers[i]);
			}

			return 0;
		}

		headers[i]=strdup(readline(sock));
		my_putline(headers[i]);

		/*
		 * If this is the end-of-request marker quit reading
		 * headers
		 */
		if (*(headers[i])=='\r' && *(headers[i]+1)=='\n') {
			free(headers[i]);
			headers[i]=0;
			break;
		}

		i++;
	}

	/*
	 * Break the request into its atomic parts
	 */
	req=breakdown_request(request);

	/*
	 * Process the request smurfed or unsmurfed, as was
	 * determined previously
	 */
	if (req->smurf_this) {
		newreq=malloc(strlen(request)+1);
		sprintf(newreq,"%s /%s %s\r\n",req->cmd,req->file,req->version);

		if (verbose_flag)
			my_printf("New request is: %s\n",newreq);

		process_smurfed(req,sock,newreq,headers);
		free(newreq);
	} else
		process_unsmurfed(req,sock,request,headers);

	close(sock);

	/*
	 * Free the request and the headers
	 */
	free(request);
	free_request(req);

	for (i=0; headers[i]; i++)
		free(headers[i]);

	/*
	 * And flush the output
	 */
	flush_output();

	return 0;
}


/*
 * Give help about how to call the program and quit
 */
void syntax() {
	puts("Syntax:\n\n");
	puts("sslsmurf [-v] [-n] [-p port] [-P proxy:port] [-h host]* [-c pem_client_cert_file -k pem_privkey_file]\n\n");
	exit(1);
}

/*
 * Process the command line options
 */
void get_options(int argc, char *argv[]) {
	char c;
	struct host_to_smurf *work;
	char *s;
	char buf[BUFSIZE];

	/*
	 * See getopt(3) to understand this :-)
	 */
	while (1==1) {
		c=getopt(argc,argv,"vnh:P:p:c:k:");

		if (c<0)
			break;

		switch (c) {
			case 'n':
				no_threading=1;
				break;

			case 'v':
				verbose_flag=1;
				break;

			case 'c':
				/*
				 * Record the client certificate file to be used
				 */
				client_cert=optarg;
				break;

			case 'k':
				/*
				 * Record the client private key file to be used
				 */
				client_privkey=optarg;
				break;

			case 'h':
				/*
				 * Add a new record to the list of hosts to be smurfed
				 */
				work=(struct host_to_smurf *)malloc(sizeof(struct host_to_smurf));
				work->host=optarg;
				work->next=root;
				root=work;
				break;

			case 'P':
				/*
				 * Take the downstream proxy to be used as host:port
				 * If the port is not specified, it is 8080
				 */
				strcpy(buf,optarg);
				proxy_host=strdup(strtok(buf,":"));
				s=strtok(0,"");

				if (!s)
					s="8080";

				proxy_port=atoi(s);

				/*
				 * If the port was not a valid number give syntax info
				 */
				if (!proxy_port)
					syntax();

				break;

			case 'p':
				/*
				 * Set the local port to accept connections ons
				 */
				port=atoi(optarg);

				if (!port)
					syntax();

				break;

			default:
				syntax();
		}
	}

	/*
	 * Info
	 */
	if (!root)
		my_puts("You have not specified hosts to smurf. Are you sure this is allright?");

	if (proxy_host)
		my_printf("Using downstream proxy %s:%d\n\n",proxy_host,proxy_port);
}

/*
 * Initialise OpenSSL
 */
void init_ssl() {
	SSL_load_error_strings();
	SSLeay_add_ssl_algorithms();

	/*
	 * Initialize a new SSLv3 context
	 */
	ctx = SSL_CTX_new (SSLv3_client_method()); 
	CHK_NULL(ctx,"SSL_CTX_new()");

	/*
	 * Set up the client certificate to be used (if so requested)
	 */
	if (client_cert && client_privkey) {
		SSL_CTX_use_certificate_file(ctx,client_cert,SSL_FILETYPE_PEM);
		SSL_CTX_use_PrivateKey_file(ctx,client_privkey,SSL_FILETYPE_PEM);

		if (!SSL_CTX_check_private_key(ctx))
			puts("Private key does not seem to match certificate!");
	}
}

/*
 * MAIN
 */
int main(int argc, char *argv[]) {
  	int 			listen_sd;
  	int 			sd;
  	struct sockaddr_in 	sa_serv;
  	struct sockaddr_in 	sa_cli;
  	size_t 			client_len;
	int			on=1;
	pthread_t		tid;
	pthread_attr_t		attr;
	struct output_chunk 	*root=0;

	/*
	 * Perform pthreads initialization
	 */
	pthread_attr_init(&attr);
	pthread_key_create(&buffer_key,0);
	pthread_key_create(&output_key,0);
  
  	/*
	 * Store the output chunk list root in the thread local storage keyspace
	 */
	if (pthread_setspecific(output_key,&root)!=0) {
		perror("Could not store output chunk list root in TLS");
		return 0;
	}

  	my_printf("This is the sslsmurf %s.\n",VERSION);
	my_puts("(c) Copyright 2004 Jos Visser (a.k.a. muppet) <josv@osp.nl>\n");

	/*
	 * Process the command line options
	 */
  	get_options(argc,argv);

	/*
	 * Initialize OpenSSL
	 */
	init_ssl();
	
	/*
	 * Create a new server socket
	 */
	listen_sd=socket(AF_INET,SOCK_STREAM,0);

	/*
	 * Reuse local address if it is still lingering
	 * in CLOSE_WAIT from a previous incarnation
	 */
	setsockopt(listen_sd,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on));
  
  	/*
	 * Set up the local address: everybody may connect
	 */
  	memset (&sa_serv, '\0', sizeof(sa_serv));
	sa_serv.sin_family=AF_INET;
  	sa_serv.sin_addr.s_addr=INADDR_ANY;
  	sa_serv.sin_port=htons(port);
  
  	/*
	 * Bind the server socket
	 */
  	if (bind(listen_sd,(struct sockaddr*)&sa_serv,sizeof (sa_serv))<0) {
		my_perror("bind() failed");
		flush_output();
		exit(1);
	}
	     
	/*
	 * The listen queue for this socket is 5 entries deep.
	 * Why 5? 
	 */
  	listen(listen_sd,5);
  	client_len=sizeof(sa_cli);

	while (!stop) {
		flush_output();

		/*
		 * Accept a new connection
		 */
  		sd=accept(listen_sd,(struct sockaddr*)&sa_cli,&client_len);

		/*
		 * If there was an error, we quit
		 */
		if (sd<0) {
			my_perror("accept() failed");
			flush_output();
			exit(1);
		}

		//my_puts("New connection accepted");
		
		/*
		 * Handle the request, either as a separate thread or
		 * not
		 */
		if (no_threading)
			process((void *)sd);
		else if (pthread_create(&tid,&attr,process,(void *)sd)!=0)
			my_perror("Could not create thread to process this request");
	}

	SSL_CTX_free (ctx);

	return 0;
}
